from .process_local_cpu_utilization import (
    ProcessLocalCPUUtilization,
    CPUStatistics,
)

# TODO(Kyle): Find a more elegant way to express these exports
__all__ = ["ProcessLocalCPUUtilization", "CPUStatistics"]



"""
Memory consumption measurement for local training processes.
"""

import time
import subprocess

from ..property import Property
from ...platform.os import is_windows


class MemoryStatistics:
    """
    The MemoryStatistics class encapsulates data
    and functionality for tracking and updating memory
    consumption statistics for a running process.
    """

    def __init__(self, avg: float, min: int, max: int):
        """
        Initialize a MemoryStatistics instance.

        :param avg: The average memory consumtion (bytes)
        :type avg: float
        :param min: The minimum memory consumption (bytes)
        :type avg: float
        :param max: The maximum memory consumption (bytes)
        :type max: float
        """
        # The statistics
        self.avg = avg
        self.min = min
        self.max = max

    def __str__(self) -> str:
        """Return a string representation of MemoryStatistics."""
        s = ""
        s += f"Average: {int(self.avg)}\n"
        s += f"Minimum: {self.min}\n"
        s += f"Maximum: {self.max}"
        return s


def _get_memory_usage(pid: int) -> int:
    """
    Get the current memory usage for the process with `pid`.
    :param pid The identifier of the process
    :return The current memory usage in KB
    """
    # sudo pmap 917 | tail -n 1 | awk '/[0-9]K/{print $2}'
    try:
        with subprocess.Popen(
            ["pmap", f"{pid}"], stdout=subprocess.PIPE
        ) as pmap, subprocess.Popen(
            ["tail", "-n", "1"], stdin=pmap.stdout, stdout=subprocess.PIPE
        ) as tail:
            used = subprocess.check_output(
                ["awk", "/[0-9]K/{print $2}"], stdin=tail.stdout
            )
        return int(used.decode("utf-8").strip()[:-1])
    except ValueError:
        return 0


class ProcessLocalMemoryConsumption(Property):
    """Measure memory consumption for a local training process."""

    def __init__(self):
        """Initialize a ProcessLocalMemoryConsumption instance."""
        super().__init__("ProcessLocalMemoryConsumption")
        if is_windows():
            raise RuntimeError(
                f"Property {self.name} is not supported on Windows."
            )

    def __call__(self, pid: int, poll_interval: int = 1) -> MemoryStatistics:
        """
        Monitor memory consumption of process at `pid` until exit.

        :param pid: The process identifier
        :type pid: int
        :param poll_interval: The poll interval, in seconds
        :type poll_interval: int

        :return The collection of memory usage statistics
        :rtype: MemoryStatistics
        """
        stats = []
        while True:
            kb = _get_memory_usage(pid)
            if kb == 0:
                break
            stats.append(kb)
            time.sleep(poll_interval)

        return MemoryStatistics(sum(stats) / len(stats), min(stats), max(stats))



"""
Superclass for all model properties.
"""


class Property:
    """
    The superclass for all model properties.
    """

    def __init__(self, name: str):
        """
        Initialize a new Property.
        :param name The name of the property
        """
        self.name = name



"""
Storage capacity measurement for locally-stored models.
"""

import os

from ..property import Property


class LocalModelSize(Property):
    """Measure the size of a locally-stored model."""

    def __init__(self):
        """Initialize a new LocalModelSize property."""
        super().__init__("LocalModelSize")

    def __call__(self, path: str) -> int:
        """
        Compute the size of the model at `path`.

        :param path: The path to the model
        :type path: str

        :return: The size of the model, in bytes
        :rtype: int
        """
        if not os.path.isfile(path) and not os.path.isdir(path):
            raise RuntimeError(f"Invalid path: {path}")

        # If the model is just a file, return it immediately
        if os.path.isfile(path):
            return os.path.getsize(path)

        # Otheriwse, the model must be directory
        assert os.path.isdir(path), "Broken invariant."

        total_size = 0
        for dirpath, _, filenames in os.walk(path):
            for name in filenames:
                path = os.path.join(dirpath, name)
                if not os.path.islink(path):
                    total_size += os.path.getsize(path)

        return total_size



"""
CPU utilization measurement for local training processes.
"""

import time
import subprocess
from subprocess import SubprocessError

from ..property import Property
from ...platform.os import is_windows


class CPUStatistics:
    """
    The CPUStatistics class encapsulates data
    and functionality for tracking and updating
    CPU consumption statistics for a running process.
    """

    def __init__(self, avg: float, min: float, max: float):
        """
        Initialize a CPUStatistics instance
        :param avg The average utilization
        :param min The minimum utilization
        :param max The maximum utilization
        """
        self.avg = avg
        self.min = min
        self.max = max

    def __str__(self) -> str:
        """Return a string representation of CPUStatistics."""
        s = ""
        s += f"Average: {self.avg:.1f}%\n"
        s += f"Minimum: {self.min:.1f}%\n"
        s += f"Maximum: {self.max:.1f}%"
        return s


def _get_cpu_usage(pid: int) -> float:
    """
    Get the current CPU usage for the process with `pid`.
    :param pid The identifier of the process
    :return The current CPU utilization as percentage
    """
    try:
        stdout = subprocess.check_output(
            ["ps", "-p", f"{pid}", "-o", "%cpu"]
        ).decode("utf-8")
        return float(stdout.strip().split("\n")[1].strip())
    except SubprocessError:
        return -1.0
    except ValueError:
        return -1.0


class ProcessLocalCPUUtilization(Property):
    """Measures CPU utilization for a local training process."""

    def __init__(self):
        """Initialize a new ProcessLocalCPUUtilization property."""
        super().__init__("ProcessLocalCPUUtilization")
        if is_windows():
            raise RuntimeError(
                f"Property {self.name} is not supported on Windows."
            )

    def __call__(self, pid: int, poll_interval: int = 1) -> CPUStatistics:
        """
        Monitor the CPU utilization of process at `pid` until exit.

        :param pid: The process identifier
        :type pid: int
        :param poll_interval: The poll interval in seconds
        :type poll_interval: int

        :return: The collection of CPU usage statistics
        :rtype: CPUStatistics
        """
        stats = []
        while True:
            util = _get_cpu_usage(pid)
            if util < 0.0:
                break
            stats.append(util)
            time.sleep(poll_interval)

        return CPUStatistics(sum(stats) / len(stats), min(stats), max(stats))



from .process_local_memory_consumption import (
    ProcessLocalMemoryConsumption,
    MemoryStatistics,
)

# TODO(Kyle): Find a more elegant way to do this
__all__ = ["ProcessLocalMemoryConsumption", "MemoryStatistics"]
